(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = global || self, factory((global.Typed = global.Typed || {}, global.Typed.Lambda = {})));
}(this, function (exports) { 'use strict';

  /**
   * Create a function that always returns a given value
   * @param value :: a
   * @returns (...* -> a)
   */
  const always = (value) => (..._) => value;

  /**
   * Allow a fixed length function to be partially applied.
   * @param f Function you'd like to curry
   */
  const curry = (f) => {
      return curriedN(f.length, f, []);
  };
  function curriedN(arity, f, previousArgs) {
      if (arity < 2) {
          return f;
      }
      return (...args) => {
          const concatArgs = previousArgs.concat(args);
          return concatArgs.length >= arity ? f(...concatArgs) : curriedN(arity, f, concatArgs);
      };
  }

  /**
   * Call a function with a list of arguments
   * @param args List of arguments
   * @param fn Function to call
   */
  const apply = curry((args, fn) => fn(...args));

  const ALL_PROPERTIES_NOT_FOUND = new Error('All Properties Not Found');
  const defaultObject = Object.freeze(Object.create(null));
  /**
   * Uses ES2015 Proxy to partially apply a function that takes in an options object
   * @param f :: Object a => (a -> b)
   */
  const curryObj = (f) => _curryObj(f, defaultObject);
  function _curryObj(f, previousObj) {
      return (x) => {
          const obj = proxyObject(Object.assign({}, previousObj, x));
          try {
              return f(obj);
          }
          catch (error) {
              if (error !== ALL_PROPERTIES_NOT_FOUND) {
                  throw error;
              }
              return _curryObj(f, obj);
          }
      };
  }
  function proxyObject(a) {
      return new Proxy(a, {
          get(target, key) {
              if (!target.hasOwnProperty(key)) {
                  throw ALL_PROPERTIES_NOT_FOUND;
              }
              return target[key];
          },
      });
  }

  /**
   * Reverse the first two arguments of a function
   * @param fn: (a -> b -> c)
   * @returns (b -> a -> c)
   */
  const flip = (fn) => ((a, b, ...args) => apply([b, a, ...args], fn));

  /**
   * Identity function
   * @param value :: a - Value to return
   */
  const id = (value) => value;

  function typeOf(value) {
      if (value === null) {
          return 'Null';
      }
      if (value === void 0) {
          return `Undefined`;
      }
      return Object.prototype.toString.call(value).slice(8, -1);
  }

  const FUNCTION_NAME_REGEX = /^function\s*([\w$]+)/;
  const DEFAULT_MATCH = ['', ''];
  /**
   * Returns the name of a function.
   * @name functionName(fn: Function): string
   */
  function functionName(fn) {
      if (fn.name) {
          return fn.name;
      }
      const [, name] = String(fn).match(FUNCTION_NAME_REGEX) || DEFAULT_MATCH;
      return name;
  }

  function includesWith(pred, x, list) {
      let idx = 0;
      const len = list.length;
      while (idx < len) {
          if (pred(x, list[idx], idx)) {
              return true;
          }
          idx += 1;
      }
      return false;
  }

  function equals(a, b, stackA = [], stackB = []) {
      if (Object.is(a, b)) {
          return true;
      }
      const typeA = typeOf(a);
      if (typeA !== typeOf(b)) {
          return false;
      }
      if (a == null || b == null) {
          return false;
      }
      switch (typeA) {
          case 'Arguments':
          case 'Array':
          case 'Object':
              if (typeof a.constructor === 'function' && functionName(a.constructor) === 'Promise') {
                  return a === b;
              }
              break;
          case 'Boolean':
          case 'Number':
          case 'String':
              if (!(typeof a === typeof b && Object.is(a.valueOf(), b.valueOf()))) {
                  return false;
              }
              break;
          case 'Date':
              if (!Object.is(a.valueOf(), b.valueOf())) {
                  return false;
              }
              break;
          case 'Error':
              return a.name === b.name && a.message === b.message;
          case 'RegExp':
              if (!(a.source === b.source &&
                  a.global === b.global &&
                  a.ignoreCase === b.ignoreCase &&
                  a.multiline === b.multiline &&
                  a.sticky === b.sticky &&
                  a.unicode === b.unicode)) {
                  return false;
              }
              break;
      }
      let idx = stackA.length - 1;
      while (idx >= 0) {
          if (stackA[idx] === a) {
              return stackB[idx] === b;
          }
          idx -= 1;
      }
      switch (typeA) {
          case 'Map':
              if (a.size !== b.size) {
                  return false;
              }
              return _uniqContentEquals(a.entries(), b.entries(), stackA.concat([a]), stackB.concat([b]));
          case 'Set':
              if (a.size !== b.size) {
                  return false;
              }
              return _uniqContentEquals(a.values(), b.values(), stackA.concat([a]), stackB.concat([b]));
          case 'Arguments':
          case 'Array':
          case 'Object':
          case 'Boolean':
          case 'Number':
          case 'String':
          case 'Date':
          case 'Error':
          case 'RegExp':
          case 'Int8Array':
          case 'Uint8Array':
          case 'Uint8ClampedArray':
          case 'Int16Array':
          case 'Uint16Array':
          case 'Int32Array':
          case 'Uint32Array':
          case 'Float32Array':
          case 'Float64Array':
          case 'ArrayBuffer':
              break;
          default:
              // Values of other types are only equal if identical.
              return false;
      }
      const keysA = Object.keys(a);
      if (keysA.length !== Object.keys(b).length) {
          return false;
      }
      const extendedStackA = stackA.concat([a]);
      const extendedStackB = stackB.concat([b]);
      idx = keysA.length - 1;
      while (idx >= 0) {
          const key = keysA[idx];
          if (!(Object.prototype.hasOwnProperty.call(b, key) &&
              equals(b[key], a[key], extendedStackA, extendedStackB))) {
              return false;
          }
          idx -= 1;
      }
      return true;
  }
  function _uniqContentEquals(aIterable, bIterable, stackA, stackB) {
      const a = Array.from(aIterable);
      const b = Array.from(bIterable);
      // tslint:disable-next-line:variable-name
      function eq(_a, _b) {
          return equals(_a, _b, stackA.slice(), stackB.slice());
      }
      // if *a* array contains any element that is not included in *b*
      // tslint:disable-next-line:no-shadowed-variable
      return !includesWith((b, aItem) => !includesWith(eq, aItem, b), b, a);
  }

  function indexOf(list, a, idx = 0) {
      // Array.prototype.indexOf doesn't exist below IE9
      if (typeof list.indexOf === 'function') {
          switch (typeof a) {
              case 'number':
                  let inf;
                  let item;
                  if (a === 0) {
                      // manually crawl the list to distinguish between +0 and -0
                      inf = 1 / a;
                      while (idx < list.length) {
                          item = list[idx];
                          if (item === 0 && 1 / item === inf) {
                              return idx;
                          }
                          idx += 1;
                      }
                      return -1;
                  }
                  else if (a !== a) {
                      // NaN
                      while (idx < list.length) {
                          item = list[idx];
                          if (typeof item === 'number' && item !== item) {
                              return idx;
                          }
                          idx += 1;
                      }
                      return -1;
                  }
                  // non-zero numbers can utilise Set
                  return list.indexOf(a, idx);
              // all these types can utilise Set
              case 'string':
              case 'boolean':
              case 'function':
              case 'undefined':
                  return list.indexOf(a, idx);
              case 'object':
                  if (a === null) {
                      // null can utilise Set
                      return list.indexOf(a, idx);
                  }
          }
      }
      while (idx < list.length) {
          if (equals(list[idx], a)) {
              return idx;
          }
          idx += 1;
      }
      return -1;
  }

  function mapArrayLike(fn, functor) {
      let idx = 0;
      const len = functor.length;
      const result = Array(len);
      while (idx < len) {
          result[idx] = fn(functor[idx]);
          idx += 1;
      }
      return result;
  }

  const STEPS = [
      [/\\/g, '\\\\'],
      [/[\b]/g, '\\b'],
      [/\f/g, '\\f'],
      [/\n/g, '\\n'],
      [/\r/g, '\\r'],
      [/\t/g, '\\t'],
      [/\v/g, '\\v'],
      [/\0/g, '\\0'],
  ];
  const LAST_STEP = [/"/g, '\\"'];
  function quote(s) {
      const escaped = STEPS.reduce(applyStep, s);
      return '"' + applyStep(escaped, LAST_STEP) + '"';
  }
  function applyStep(str, step) {
      return str.replace(step[0], step[1]);
  }

  /**
   * Convert anything into a string
   * @param x :: any
   * @returns a string representation of a value
   */
  const toString = (x) => _toString(x, []);
  function _toString(x, seen) {
      const recur = (y) => {
          const xs = seen.concat([x]);
          return indexOf(xs, y) > -1 ? '<Circular>' : _toString(y, xs);
      };
      const mapPairs = (obj, keys) => {
          return mapArrayLike((k) => {
              return quote(k) + ': ' + recur(obj[k]);
          }, keys.slice().sort());
      };
      switch (Object.prototype.toString.call(x)) {
          case '[object Arguments]':
              return '(function() { return arguments; }(' + mapArrayLike(recur, x).join(', ') + '))';
          case '[object Array]':
              return ('[' +
                  mapArrayLike(recur, x)
                      .concat(mapPairs(x, Object.keys(x).filter(k => !/^\d+$/.test(k))))
                      .join(', ') +
                  ']');
          case '[object Boolean]':
              return typeof x === 'object' ? 'new Boolean(' + recur(x.valueOf()) + ')' : x.toString();
          case '[object Date]':
              return ('new Date(' + (isNaN(x.valueOf()) ? recur(NaN) : quote(x.toISOString())) + ')');
          case '[object Null]':
              return 'null';
          case '[object Number]':
              return typeof x === 'object'
                  ? 'new Number(' + recur(x.valueOf()) + ')'
                  : 1 / x === -Infinity
                      ? '-0'
                      : x.toString(10);
          case '[object String]':
              return typeof x === 'object' ? 'new String(' + recur(x.valueOf()) + ')' : quote(x);
          case '[object Undefined]':
              return 'undefined';
          default:
              if (typeof x.toString === 'function') {
                  const repr = x.toString();
                  if (repr !== '[object Object]') {
                      return repr;
                  }
              }
              return '{' + mapPairs(x, Object.keys(x)).join(', ') + '}';
      }
  }

  /**
   * Memoize a function
   * @param f Function to memoize
   */
  const memoize = (f) => {
      const cache = new Map();
      return (...args) => {
          const key = args.reduce((x, y) => x + toString(y), '');
          if (cache.has(key)) {
              return cache.get(key);
          }
          let result = f(...args);
          if (typeof result === 'function') {
              result = memoize(result);
          }
          cache.set(key, result);
          return result;
      };
  };

  /**
   * Function that does nothing
   */
  const noOp = () => void 0;

  /** Left-to-right composition for exactly two function */
  const pipe2 = (f, g) => (x) => g(f(x));
  /**
   * Generic Left-to-right composition
   */
  const pipe = ((...fns) => fns.length > 1 ? fns.slice(1).reduce(pipe2, fns[0]) : fns[0]);

  /**
   * Perform a side-effect with a value and return the given value.
   * @param fn :: (a -> *)
   * @param value :: a
   * @returns a
   */
  const tap = curry((fn, value) => {
      fn(value);
      return value;
  });

  /**
   * Convert a function like (a -> b -> c -> d) into ((a, b, c) -> d)
   * @param fn :: Function to uncurry
   * @returns Function that accepts all arguments at once.
   */
  function uncurry(f) {
      if (typeof f !== 'function' || f.length === 0) {
          return f;
      }
      // tslint:disable-next-line:only-arrow-functions
      return function () {
          let r = f;
          // tslint:disable-next-line:prefer-for-of
          for (let i = 0; i < arguments.length; i++) {
              r = r(arguments[i]);
          }
          return uncurry(r);
      };
  }

  exports.always = always;
  exports.apply = apply;
  exports.curry = curry;
  exports.curryObj = curryObj;
  exports.flip = flip;
  exports.id = id;
  exports.memoize = memoize;
  exports.noOp = noOp;
  exports.pipe = pipe;
  exports.pipe2 = pipe2;
  exports.tap = tap;
  exports.uncurry = uncurry;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=index.js.map
